'''
Copyright (C) 2021  Alexander "PapaTutuWawa"

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
# TODO: Replace most of these with a query API
import os
import json

from mira.storage import StorageManager

def append_or_insert(dict_, key, value):
    if key in dict_:
        dict_[key].append(value)
    else:
        dict_[key] = [value]

class SubscriptionManager:
    '''
    This class is tasked with providing functions that simplify dealing
    with subscriptions.

    NOTE: This class must only by instanciated *after* the StorageManager
          has been instanciated at least once. For modules, this is no
          issue as they're only created after the manager classes are
          ready.
    '''
    __instance = None

    @staticmethod
    def get_instance():
        if not SubscriptionManager.__instance:
            SubscriptionManager()
        return SubscriptionManager.__instance
    
    def __init__(self):
        self._sm = StorageManager.get_instance()
        # Module -> JID -> Keywords
        self._subscriptions = self._sm.get_data('_SubscriptionManager', 'subscriptions')

        SubscriptionManager.__instance = self

    def get_subscriptions_for(self, module, jid):
        '''
        Returns a dictionary keyword -> data which represents
        every subscription a jid has in the context of @module.
        '''
        if not module in self._subscriptions:
            return []
        if not jid in self._subscriptions[module]:
            return []

        return self._subscriptions[module][jid]

    def get_subscriptions_for_keyword(self, module, keyword):
        '''
        Returns an array of JIDs that are subscribed to the keyword of module
        '''
        if not module in self._subscriptions:
            return []

        tmp = []
        for jid in self._subscriptions[module]:
            if not keyword in self._subscriptions[module][jid]:
                continue

            data = self._subscriptions[module][jid][keyword]['data']
            tmp.append((jid, data))
        return tmp

    def get_subscriptions_for_keywords(self, module, keywords):
        '''
        Returns an array of JIDs that are subscribed to at least one of the keywords
        of module
        '''
        if not module in self._subscriptions:
            return []

        tmp = []
        keyword_set = set(keywords)
        for jid in self._subscriptions[module]:
            if set(self._subscriptions[module][jid].keys()) & keyword_set:
                continue

            data = self._subscriptions[module][jid][keyword]['data']
            tmp.append((jid, data))
        return tmp
    
    def get_subscription_keywords(self, module):
        '''Returns a list of subscribed keywords in module'''
        if not module in self._subscriptions:
            return []

        tmp = []
        for subscription in self._subscriptions[module].values():
            tmp += list(subscription.keys())
        return tmp
    
    def is_subscribed_to(self, module, jid, keyword):
        '''
        Returns True if @jid is subscribed to @keyword within the context
        of @module. False otherwise
        '''
        return keyword in self.get_subscriptions_for(module, jid)

    def is_subscribed_to_data(self, module, jid, keyword, item):
        '''
        Returns True if @jid is subscribed to the item @item inside
        the keyword @keyword within the context of @module
        '''
        subscriptions = self.get_subscriptions_for(module, jid)
        if not subscriptions:
            return False
        if not keyword in subscriptions:
            return False

        return item in subscriptions[keyword]['data']

    def is_subscribed_to_data_one(self, module, jid, keyword, func):
        '''
        Like is_subscribed_to_data, but returns True if there is at
        least one item for which func returns True.
        '''
        subscriptions = self.get_subscriptions_for(module, jid)
        if not subscriptions:
            return False
        if not keyword in subscriptions:
            return False

        for item in subscriptions[keyword]['data']:
            if func(item):
                return True
        return False
    
    def add_subscription_for(self, module, jid, keyword, data={}):
        '''
        Adds a subscription to @keyword with data @data for @jid within
        the context of @module.
        '''
        if not module in self._subscriptions:
            self._subscriptions[module] = {}
        if not jid in self._subscriptions[module]:
            self._subscriptions[module][jid] = {}

        self._subscriptions[module][jid][keyword] = {
            'data': data
        }
        self.__flush()

    def append_data_for_subscription(self, module, jid, keyword, item):
        '''
        Special helper function which appends item to the data field of
        a subscription to @keyword from @jid within the context of @module.

        If no subscription exists, then one will be created.

        NOTE: This function expects data to be a dict with key 'data', which
              must be an array, so it will fail if add_subscription_for has
              been called beforehand with data equal to anything but an dict
              with a 'data' key containing an array.
        '''
        if not module in self._subscriptions:
            self._subscriptions[module] = {}
        if not jid in self._subscriptions[module]:
            self._subscriptions[module][jid] = {}
        if not keyword in self._subscriptions[module][jid]:
            self._subscriptions[module][jid][keyword] = {
                'data': [item]
            }
            self.__flush()
            return

        self._subscriptions[module][jid][keyword]['data'].append(item)
        self.__flush()

    def remove_subscription_for(self, module, jid, keyword):
        '''
        Removes a subscription to @keyword for @jid within the context
        of @module
        '''
        del self._subscriptions[module][jid][keyword]

        if not self._subscriptions[module][jid]:
            del self._subscriptions[module][jid]
        if not self._subscriptions[module]:
            del self._subscriptions[module]

        self.__flush()

    def remove_item_for_subscription(self, module, jid, keyword, item, flush=True):
        '''
        The deletion counterpart of append_data_for_subscription.
        '''
        self.filter_items_for_subscription(module,
                                           jid,
                                           keyword,
                                           func=lambda x: x == item,
                                           flush=flush)

    def filter_items_for_subscription(self, module, jid, keyword, func, flush=True):
        '''
        remove_item_for_subscription but for multiple items
        '''
        if not module in self._subscriptions:
            return
        if not jid in self._subscriptions[module]:
            return
        if not keyword in self._subscriptions[module][jid]:
            return

        self._subscriptions[module][jid][keyword]['data'] = list(filter(func,
                                                                        self._subscriptions[module][jid][keyword]['data']))

        if not self._subscriptions[module][jid][keyword]['data']:
            del self._subscriptions[module][jid][keyword]
        if not self._subscriptions[module][jid]:
            del self._subscriptions[module][jid]
        if not self._subscriptions[module]:
            del self._subscriptions[module]

        if flush:
            self.__flush()
        
    def __flush(self):
        '''
        Write subscription data to disk. Just an interface to StorageManager
        '''
        self._sm.set_data('_SubscriptionManager', 'subscriptions', self._subscriptions)
